"""
Adapted from toil.batchSystems.lsfHelper.

https://github.com/DataBiosphere/toil/blob/master/src/toil/batchSystems/lsfHelper.py.
"""
import base64
import json
import os
import random
import subprocess
import time

from toil.lib.conversions import convert_units
from toil.batchSystems.lsf import logger
from toil.batchSystems.lsfHelper import per_core_reservation


_RESOURCES_START_TAG = "__rsrc"
_RESOURCES_CLOSE_TAG = "rsrc__"

try:
    MAX_MEMORY = int(os.getenv("TOIL_CONTAINER_RETRY_MEM", "60")) * 1e9
    MAX_RUNTIME = int(os.getenv("TOIL_CONTAINER_RETRY_RUNTIME", "40000"))
except ValueError:  # pragma: no cover
    MAX_MEMORY = 60 * 1e9
    MAX_RUNTIME = 40000
    logger.error("Failed to parse default values for resource retry.")


def _parse_memory(mem: float) -> str:
    """Parse memory parameter."""
    megabytes_of_mem = max(
        convert_units(
            float(mem),
            src_unit="B",  # Expect always bytes in `memory=<int>` if no explicit unit.
            dst_unit="MB",
        ),
        1.0,
    )
    # round as a string here to avoid returning something like 1.231e+12
    return f"{megabytes_of_mem:.0f}MB"


def encode_dict(dictionary):
    """Encode `dictionary` in string."""
    if dictionary:
        return (
            f"{_RESOURCES_START_TAG}"
            f"{base64.b64encode(json.dumps(dictionary).encode()).decode()}"
            f"{_RESOURCES_CLOSE_TAG}"
        )
    return ""


def decode_dict(string):
    """Get dictionary encoded in `string` by `encode_dict`."""
    if isinstance(string, str):
        split = string.split(_RESOURCES_START_TAG, 1)[-1]
        split = split.split(_RESOURCES_CLOSE_TAG, 1)
        if len(split) == 2:
            return json.loads(base64.b64decode(split[0]))
    return {}


def with_retries(operation, *args, **kwargs):
    """Add a random sleep after each retry."""
    latest_err = Exception

    for i in [2, 3, 5, 10]:
        try:
            return operation(*args, **kwargs)
        except subprocess.CalledProcessError as err:
            time.sleep(i + random.uniform(-i, i))
            latest_err = err
            logger.error(
                "Operation %s failed with code %d: %s",
                operation,
                err.returncode,
                err.output,
            )
    raise latest_err  # pragma: no cover


def build_bsub_line(cpu, mem, runtime, jobname, stdoutfile=None, stderrfile=None):
    """
    Build an args list for a bsub submission.

    Arguments:
        cpu (int): number of cores needed.
        mem (float): number of bytes of memory needed.
        runtime (int): estimated run time for the job in minutes.
        jobname (str): the job name.
        stdoutfile (str): filename to direct job stdout
        stderrfile (str): filename to direct job stderr

    Returns:
        list: bsub command.
    """
    bsubline = [
        "bsub",
        "-cwd",
        ".",
        "-o",
        stdoutfile or "/dev/null",
        "-e",
        stderrfile or "/dev/null",
        "-J",
        f"'{jobname}'",
    ]
    cpu = int(cpu) or 1
    if mem:
        per_core = os.getenv("TOIL_CONTAINER_LSF_PER_CORE")
        if (per_core == "Y") or (not per_core and per_core_reservation()):
            mem = mem / cpu

        mem_resource = _parse_memory(mem)
        mem_limit = _parse_memory(mem)

        bsubline += ["-R", f"select[mem>{mem_resource}]"]
        bsubline += ["-R", f"rusage[mem={mem_resource}]"]
        bsubline += ["-M", str(mem_limit)]

    if cpu:
        bsubline += ["-n", str(cpu)]

    if runtime:
        bsubline += [os.getenv("TOIL_CONTAINER_RUNTIME_FLAG", "-W"), str(int(runtime))]

    if os.getenv("TOIL_LSF_ARGS"):
        bsubline.extend(os.getenv("TOIL_LSF_ARGS").split())

    # log to lsf
    logger.info("Submitting to LSF with: %s", " ".join(bsubline))
    return bsubline



"""toil_container version test."""
import subprocess

from toil_container import utils


def test_which_util():
    """Test to check the which util is working."""
    std_out = subprocess.check_output(
        [utils.which("python"), "--version"], stderr=subprocess.STDOUT
    )
    assert std_out.decode().startswith("Python 3")



"""toil_container parsers tests."""

import click
import pytest

from toil_container import exceptions
from toil_container import parsers

from .utils import DOCKER_IMAGE
from .utils import SINGULARITY_IMAGE
from .utils import SKIP_DOCKER
from .utils import SKIP_SINGULARITY
from .utils import Capturing


def check_help_toil(parser):
    with Capturing() as plain:
        try:
            parser.parse_args(["--help"])
        except SystemExit:
            pass

    with Capturing() as with_toil:
        try:
            parser.parse_args(["--help-toil"])
        except SystemExit:
            pass

    plain = "\n".join(plain)
    with_toil = "\n".join(with_toil)

    # by default toil options shouldn't be printed
    assert "Toil core options" not in plain
    assert "toil arguments" in plain

    # check that toil options were added by default
    assert "Toil core options" in with_toil
    assert "toil arguments" in with_toil

    # test container arguments
    if isinstance(parser, parsers.ContainerArgumentParser):
        with Capturing() as with_container:
            try:
                parser.parse_args(["--help-container"])
            except SystemExit:
                pass

        with_container = "\n".join(with_container)
        assert "container arguments:" not in plain
        assert "container arguments:" in with_container


def assert_parser_volumes(image_flag, image, tmpdir):
    # test valid volumes
    args = [
        image_flag,
        image,
        "--volumes",
        tmpdir.mkdir("vol1").strpath,
        "/vol1",
        "--volumes",
        tmpdir.mkdir("vol2").strpath,
        "/vol2",
        "--workDir",
        tmpdir.mkdir("workDir").strpath,
        "jobstore",
    ]

    assert parsers.ContainerArgumentParser().parse_args(args)

    # test invalid volumes, dst volume is not an absolute path
    args = [
        "--volumes",
        tmpdir.join("vol1").strpath,
        "vol1",
        image_flag,
        image,
        "jobstore",
    ]

    with pytest.raises(exceptions.ValidationError):
        parsers.ContainerArgumentParser().parse_args(args)


def test_help_toil():
    check_help_toil(parsers.ToilShortArgumentParser())


def test_help_toil_container_parser():
    parser = parsers.ContainerArgumentParser()
    check_help_toil(parser)
    assert "container arguments" in parser.format_help()


def test_parser_add_version():
    parser = parsers.ToilBaseArgumentParser(version="foo")
    assert "version" in parser.format_help()


def test_container_parser_cant_use_docker_and_singularity_together():
    with pytest.raises(click.UsageError) as error:
        args = ["--singularity", "i", "--docker", "j", "jobstore"]
        parsers.ContainerArgumentParser().parse_args(args)

    assert "use --singularity or --docker, not both." in str(error.value)


def test_volumes_only_used_with_containers():
    with pytest.raises(click.UsageError) as error:
        args = ["--volumes", "foo", "bar", "jobstore"]
        parsers.ContainerArgumentParser().parse_args(args)

    assert "--volumes should be used only " in str(error.value)


@SKIP_DOCKER
def test_container_parser_docker_valid_image():
    args = ["--docker", DOCKER_IMAGE, "jobstore"]
    assert parsers.ContainerArgumentParser().parse_args(args)


@SKIP_DOCKER
def test_container_parser_docker_invalid_image():
    with pytest.raises(exceptions.ValidationError):
        args = ["--docker", "florentino-ariza-img", "jobstore"]
        assert parsers.ContainerArgumentParser().parse_args(args)


@SKIP_DOCKER
def test_container_parser_docker_volumes(tmpdir):
    assert_parser_volumes("--docker", DOCKER_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_container_parser_singularity_valid_image():
    args = ["--singularity", SINGULARITY_IMAGE, "jobstore"]
    assert parsers.ContainerArgumentParser().parse_args(args)


@SKIP_SINGULARITY
def test_container_parser_singularity_invalid_image():
    with pytest.raises(exceptions.ValidationError):
        args = ["--singularity", "florentino-ariza-img", "jobstore"]
        assert parsers.ContainerArgumentParser().parse_args(args)


@SKIP_SINGULARITY
def test_container_parser_singularity_volumes(tmpdir):
    assert_parser_volumes("--singularity", SINGULARITY_IMAGE, tmpdir)



"""toil_container jobs tests."""

import os
import subprocess
import time

from toil.batchSystems.lsfHelper import per_core_reservation

from toil_container import parsers
from toil_container import jobs
from toil_container import lsf_helper

from .utils import SKIP_LSF


TEST_QUEUE = "general"


class testJobRuntimeRetry(jobs.ContainerJob):
    def run(self, fileStore):
        time.sleep(70)


def test_with_retries(tmpdir):
    test_path = tmpdir.join("test")

    def _test():
        if not os.path.isfile(test_path.strpath):
            test_path.write("hello")
            subprocess.check_call(["rm", "/florentino-ariza-volume"])
        return

    lsf_helper.with_retries(_test)


def test_encode_decode_resources():
    expected = {"runtime": 1}
    e_string = lsf_helper.encode_dict(expected)
    obtained = lsf_helper.decode_dict(e_string)
    assert lsf_helper.encode_dict({}) == ""
    assert lsf_helper.decode_dict("") == {}
    assert expected == obtained


def test_build_bsub_line():
    os.environ["TOIL_LSF_ARGS"] = f"-q {TEST_QUEUE}"
    os.environ["TOIL_CONTAINER_LSF_PER_CORE"] = "N"
    mem = 2147483648
    cpu = 1

    obtained = lsf_helper.build_bsub_line(
        cpu=cpu, mem=mem, runtime=1, jobname="Test Job"
    )
    mem_resource = lsf_helper._parse_memory(mem)
    mem_limit = lsf_helper._parse_memory(mem)

    expected = [
        "bsub",
        "-cwd",
        ".",
        "-o",
        "/dev/null",
        "-e",
        "/dev/null",
        "-J",
        "'Test Job'",
        "-R",
        f"select[mem>{mem_resource}]",
        "-R",
        f"rusage[mem={mem_resource}]",
        "-M",
        str(mem_limit),
        "-n",
        "1",
        "-W",
        "1",
        "-q",
        TEST_QUEUE,
    ]

    assert obtained == expected
    del os.environ["TOIL_LSF_ARGS"]
    del os.environ["TOIL_CONTAINER_LSF_PER_CORE"]


@SKIP_LSF
def test_build_bsub_line_zero_cpus():
    command = lsf_helper.build_bsub_line(
        cpu=0,
        mem=2147483648,
        runtime=1,
        jobname="Test Job",
    )
    assert "-n 1" in " ".join(command)


@SKIP_LSF
def test_bsubline_works():
    command = lsf_helper.build_bsub_line(
        cpu=1, mem=2147483648, runtime=1, jobname="Test Job"
    )
    command.extend(["-K", "echo"])
    assert subprocess.check_call(command) == 0


@SKIP_LSF
def test_custom_lsf_batch_system(tmpdir):
    jobstore = tmpdir.join("jobstore").strpath
    log = tmpdir.join("log.txt").strpath
    options = parsers.ToilBaseArgumentParser().parse_args([jobstore, "--logFile", log])
    options.batchSystem = "custom_lsf"
    options.disableCaching = True
    job = jobs.ContainerJob(options, runtime=1)
    jobs.ContainerJob.Runner.startToil(job, options)

    with open(log, encoding="utf-8") as f:
        assert "-W 1" in f.read()


@SKIP_LSF
def test_custom_lsf_resource_retry_runtime(tmpdir):
    jobstore = tmpdir.join("jobstore").strpath
    log = tmpdir.join("log.txt").strpath
    options = parsers.ToilBaseArgumentParser().parse_args([jobstore, "--logFile", log])
    options.batchSystem = "custom_lsf"
    options.disableCaching = True
    job = testJobRuntimeRetry(options, runtime=1)
    testJobRuntimeRetry.Runner.startToil(job, options)

    with open(log, encoding="utf-8") as f:
        assert "Detected job killed by LSF" in f.read()


@SKIP_LSF
def test_custom_lsf_per_core_env(tmpdir):
    jobstore = tmpdir.join("jobstore").strpath
    log = tmpdir.join("log.txt").strpath
    options = parsers.ToilBaseArgumentParser().parse_args([jobstore, "--logFile", log])
    options.batchSystem = "custom_lsf"
    options.disableCaching = True

    # Set total memory per job
    os.environ["TOIL_CONTAINER_LSF_PER_CORE"] = "N"
    job_1 = jobs.ContainerJob(options, runtime=1, cores=2, memory="10G")
    jobs.ContainerJob.Runner.startToil(job_1, options)
    with open(log, "rt", encoding="utf-8") as f:
        assert "-M 10000MB -n 2" in f.read()

    # Set total memory per core
    os.environ["TOIL_CONTAINER_LSF_PER_CORE"] = "Y"
    job_2 = jobs.ContainerJob(options, runtime=1, cores=2, memory="10G")
    jobs.ContainerJob.Runner.startToil(job_2, options)
    with open(log, "rt", encoding="utf-8") as f:
        assert "-M 5000MB -n 2" in f.read()

    # Use per_core_reservation() from lsf config
    del os.environ["TOIL_CONTAINER_LSF_PER_CORE"]
    job_3 = jobs.ContainerJob(options, runtime=1, cores=2, memory="10G")
    jobs.ContainerJob.Runner.startToil(job_3, options)
    with open(log, "rt", encoding="utf-8") as f:
        assert (
            "-M 5000MB -n 2" if per_core_reservation() else "-M 10000MB -n 2"
        ) in f.read()



"""toil_container pasers."""

import argparse

from toil.job import Job
import click

from toil_container import validators

SHOW_TOILGROUPS_PROPERTY = "_show_toil_groups"
SHOW_CONTGROUPS_PROPERTY = "_show_container_groups"
CUSTOM_TOIL_ACTIONS = [
    argparse.Action(
        [],
        dest="jobStore",
        help="the location of the job store for the workflow. "
        "See --help-toil for more toil options and information [REQUIRED]",
    )
]


class _ToilHelpAction(argparse._HelpAction):
    def __call__(self, parser, namespace, values, option_string=None):
        """Print parser help and exist whilst adding a flag to the parser."""
        setattr(parser, SHOW_TOILGROUPS_PROPERTY, True)
        parser.print_help()
        parser.exit()


class _ContainerHelpAction(argparse._HelpAction):
    def __call__(self, parser, namespace, values, option_string=None):
        """Print parser help and exist whilst adding a flag to the parser."""
        setattr(parser, SHOW_CONTGROUPS_PROPERTY, True)
        parser.print_help()
        parser.exit()


class ToilBaseArgumentParser(argparse.ArgumentParser):

    """Add toil options to argument parser."""

    def __init__(self, version=None, **kwargs):
        """
        Add Toil options to parser.

        Arguments:
            version (str): optionally add a version argument.
            kwargs (dict): argparse.ArgumentParser key word arguments.
        """
        if not kwargs.get("formatter_class"):
            kwargs["formatter_class"] = argparse.ArgumentDefaultsHelpFormatter

        super().__init__(**kwargs)

        if version:
            self.add_argument(
                "-v", "--version", action="version", version="%(prog)s " + str(version)
            )

        Job.Runner.addToilOptions(self)


class ToilShortArgumentParser(ToilBaseArgumentParser):

    """
    A parser with only required toil args in --help with --help-toil option.

    Toil options are automatically added, but hidden by default in the help
    print. However, the `--help-toil` argument prints toil full rocketry.
    """

    def __init__(self, **kwargs):
        """Add Toil`--help-toil` to parser."""
        super().__init__(**kwargs)

        self.add_argument(
            "--help-toil",
            action=_ToilHelpAction,
            default=argparse.SUPPRESS,
            help="show help with toil arguments and exit",
        )

    @property
    def custom_actions(self):
        """List of custom actions for usage summary."""
        return CUSTOM_TOIL_ACTIONS if not self.show_toil_groups else []

    @property
    def show_toil_groups(self):
        """Whether toil help groups should be displayed."""
        return getattr(self, SHOW_TOILGROUPS_PROPERTY, False)

    def hide_action_group(self, action_group):
        """Determine if an action group should be hidden."""
        is_toil_group = action_group.title.lower().startswith("toil")
        if is_toil_group or "Logging Options" in action_group.title:
            return not self.show_toil_groups

        return False

    def get_help_groups(self):
        """Decide whether to show toil options or not."""
        action_groups = []
        actions = []

        for action_group in self._action_groups:
            if not self.hide_action_group(action_group):
                action_groups.append(action_group)
                actions += action_group._group_actions

        return actions, action_groups

    def format_help(self):
        """Include toil options if `self.show_toil_groups` is True."""
        formatter = self._get_formatter()

        # decide whether to show toil options or not
        actions, action_groups = self.get_help_groups()

        # usage, the CUSTOM_TOIL_ACTIONS are just for display
        formatter.add_usage(
            self.usage, actions + self.custom_actions, self._mutually_exclusive_groups
        )

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        for action_group in action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # add custom toil section
        if not self.show_toil_groups:
            formatter.start_section("toil arguments")
            formatter.add_arguments(CUSTOM_TOIL_ACTIONS)
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class ContainerArgumentParser(ToilShortArgumentParser):

    """Toil Argument Parser with options for containerized system calls."""

    _ARGUMENT_GROUP_NAME = "container arguments"

    def __init__(self, *args, **kwargs):
        """Add container options to parser."""
        super().__init__(*args, **kwargs)
        settings = self.add_argument_group(self._ARGUMENT_GROUP_NAME)

        settings.add_argument(
            "--docker",
            help="name/path of the docker image available in daemon",
            default=None,
            required=False,
        )

        settings.add_argument(
            "--singularity",
            help="name/path of the singularity image available in deamon",
            default=None,
            required=False,
        )

        settings.add_argument(
            "--volumes",
            help="tuples of (local path, absolute container path)",
            required=False,
            default=None,
            action="append",
            nargs=2,
        )

        self.add_argument(
            "--help-container",
            action=_ContainerHelpAction,
            default=argparse.SUPPRESS,
            help="show help with container arguments and exit",
        )

    def hide_action_group(self, action_group):
        """Return falsey if group shouldn't be showed."""
        if action_group.title.startswith(self._ARGUMENT_GROUP_NAME):
            return not getattr(self, SHOW_CONTGROUPS_PROPERTY, False)

        return super().hide_action_group(action_group)

    def parse_args(self, args=None, namespace=None):
        """Validate parsed options."""
        args = super().parse_args(args=args, namespace=namespace)

        images = [args.docker, args.singularity]

        if all(images):
            raise click.UsageError("use --singularity or --docker, not both.")

        if args.volumes and not any(images):
            raise click.UsageError(
                "--volumes should be used only with " "--singularity or --docker."
            )

        if any(images):
            validate_kwargs = {}

            if args.volumes:
                validate_kwargs["volumes"] = args.volumes

            if args.workDir:
                validate_kwargs["working_dir"] = args.workDir

            if args.docker:
                validate_kwargs["image"] = args.docker
                validators.validate_docker(**validate_kwargs)
            else:
                validate_kwargs["image"] = args.singularity
                validators.validate_singularity(**validate_kwargs)

        return args



"""toil_container integration tests."""

from os.path import join

import pytest
from toil.leader import FailedJobsException

from toil_container import jobs
from toil_container import parsers
from toil_container.containers import _TMP_PREFIX
from .utils import DOCKER_IMAGE
from .utils import SINGULARITY_IMAGE
from .utils import SKIP_DOCKER
from .utils import SKIP_SINGULARITY


class ContainerTestJob(jobs.ContainerJob):

    """Test Job Class for Integration Test."""

    cmd = ["pwd"]
    cwd = None
    env = {}
    check_output = True

    def run(self, _):
        """Run job logic."""
        self.call(self.cmd, cwd=self.cwd, env=self.env, check_output=self.check_output)


def assert_pipeline(image_flag, image, tmpdir, parallel=True):
    """
    Make sure parallel jobs work.

                head
                |  |
          child_a  child_b
                |  |
                tail

    Make sure a failing job output is sent to the main pipeline logs

                head
                  |
               child_c
    """
    jobstore = tmpdir.join("jobstore")
    workdir = tmpdir.mkdir("working_dir")
    local_volume = tmpdir.mkdir("volume")
    container_volume = "/volume"

    # set parser
    parser = parsers.ContainerArgumentParser()
    args = [jobstore.strpath, "--workDir", workdir.strpath]

    if image_flag:
        args += [image_flag, image, "--volumes", local_volume.strpath, container_volume]

    # set testing variables
    out_file = "bottle.txt"
    tmp_file_local = workdir.join(out_file)
    tmp_file_container = join("/tmp", out_file)
    vol_file_local = local_volume.join(out_file)
    vol_file_container = join(container_volume, out_file)

    if image_flag is None:
        tmp_file_container = tmp_file_local.strpath
        vol_file_container = vol_file_local.strpath

    # create jobs
    options = parser.parse_args(args)

    if parallel:
        # Run the parallel pipeline
        head = ContainerTestJob(options)
        child_a = ContainerTestJob(options)
        child_b = ContainerTestJob(options)
        tail = ContainerTestJob(options)

        # assign commands and attributes
        cmd = ["/bin/bash", "-c"]

        # test cwd and workDir, _rm_tmp_dir is used to prevent tmpdir to be removed
        head._rm_tmp_dir = False
        head.cwd = "/bin"
        head.cmd = cmd + ["pwd >> " + tmp_file_container]

        # test env
        child_a.env = {"FOO": "BAR"}
        child_a.cmd = cmd + ["echo $FOO >> " + vol_file_container]

        # test check_output
        child_b.check_output = False
        child_b.cmd = cmd + ["echo check_call >> " + vol_file_container]

        # test volumes
        tail.cmd = cmd + ["echo volume >> " + vol_file_container]

        # build dag
        head.addChild(child_a)
        head.addChild(child_b)
        head.addFollowOn(tail)

        # start pipeline
        jobs.ContainerJob.Runner.startToil(head, options)

        if image_flag:
            pattern = join(_TMP_PREFIX + "*", out_file)
            tmp_file_local = next(workdir.visit(pattern))

        # Test the output
        with open(tmp_file_local.strpath) as f:
            result = f.read()
            assert "/bin" in result

        if image_flag:
            with open(vol_file_local.strpath) as f:
                result = f.read()
                assert "volume" in result
                assert "BAR" in result
                assert "check_call" in result

    else:
        # Run the failing pipeline
        options.retryCount = 0
        head = ContainerTestJob(options)
        child_c = ContainerTestJob(options)
        child_c.cmd = ["rm", "/florentino-arisa"]

        head.addChild(child_c)

        with pytest.raises(FailedJobsException) as captured_error:
            # start pipeline
            jobs.ContainerJob.Runner.startToil(head, options)

        assert (
            "rm: cannot remove '/florentino-arisa': No such file or directory"
        ) in captured_error.value.msg


def test_pipeline_with_subprocess(tmpdir):
    """Run the parallel pipeline."""
    assert_pipeline(None, None, tmpdir)


def test_failing_pipeline_with_subprocess(tmpdir):
    """Run the failing pipeline."""
    assert_pipeline(None, None, tmpdir, False)


@SKIP_DOCKER
def test_pipeline_with_docker(tmpdir):
    assert_pipeline("--docker", DOCKER_IMAGE, tmpdir)


@SKIP_SINGULARITY
def test_pipeline_with_singularity(tmpdir):
    assert_pipeline("--singularity", SINGULARITY_IMAGE, tmpdir)



"""A custom LSF batchsystem to process additional resources."""
# pylint: disable=C0103, W0223

from collections import defaultdict
from datetime import datetime
import os
import subprocess

from toil.batchSystems.lsf import LSFBatchSystem, logger
from toil.batchSystems.abstractBatchSystem import UpdatedBatchJobInfo

from toil_container.lsf_helper import (
    MAX_MEMORY,
    MAX_RUNTIME,
    build_bsub_line,
    decode_dict,
    with_retries,
)


class CustomLSFBatchSystem(LSFBatchSystem):

    """A custom LSF batchsystem used to encode extra lsf resources."""

    def __init__(self, *args, **kwargs):
        """Create a mapping table for JobIDs to JobNodes."""
        super().__init__(*args, **kwargs)
        self.Id2Node = {}
        self.resourceRetryCount = defaultdict(set)

    def issueBatchJob(self, jobDesc, job_environment=None):
        """Load the jobDesc into the JobID mapping table."""
        jobID = super().issueBatchJob(jobDesc, job_environment)
        self.Id2Node[jobID] = jobDesc
        return jobID

    @staticmethod
    def with_retries(operation, *args, **kwargs):
        """Add a random sleep after each retry."""
        return with_retries(operation, *args, **kwargs)

    class Worker(LSFBatchSystem.Worker):

        """Make prepareBsub a class method and parse unitName."""

        _CANT_DETERMINE_JOB_STATUS = "NO STATUS FOUND"

        def forgetJob(self, jobID):
            """Remove jobNode from the mapping table when forgetting."""
            self.boss.Id2Node.pop(jobID, None)
            self.boss.resourceRetryCount.pop(jobID, None)
            return super().forgetJob(jobID)

        def prepareBsub(self, cpu, mem, jobID, runtime=None):  # pylint: disable=W0221
            """
            Make a bsub commandline to execute.

            Arguments:
                cpu (int): number of cores needed.
                mem (float): number of bytes of memory needed.
                jobID (str): ID number of the job.
                runtime (int): total runtime.

            Returns:
                list: a bsub line argument.
            """
            env_jobname = os.getenv("TOIL_LSF_JOBNAME", "Toil Job")

            try:  # try to update runtime if not provided
                jobNode = self.boss.Id2Node[jobID]
                runtime = runtime or decode_dict(jobNode.unitName).get("runtime", None)
                jobname = f"{env_jobname} {jobNode.jobName} {jobID}"
            except KeyError:
                jobname = f"{env_jobname} {jobID}"

            stdoutfile: str = self.boss.formatStdOutErrPath(jobID, "%J", "out")
            stderrfile: str = self.boss.formatStdOutErrPath(jobID, "%J", "err")

            return build_bsub_line(
                cpu=cpu,
                mem=mem,
                runtime=runtime,
                jobname=jobname,
                stdoutfile=stdoutfile,
                stderrfile=stderrfile,
            )

        def checkOnJobs(self):
            """
            Check and update status of all running jobs.

            Respects statePollingWait and will return cached results if not within
            time period to talk with the scheduler.
            """
            if (
                self._checkOnJobsTimestamp
                and (datetime.now() - self._checkOnJobsTimestamp).total_seconds()
                < self.boss.config.statePollingWait
            ):
                return self._checkOnJobsCache

            activity = False
            not_finished = self.boss.with_retries(self._getNotFinishedIDs)  # Added

            for jobID in list(self.runningJobs):
                batchJobID = self.getBatchSystemID(jobID)

                if int(batchJobID) in not_finished:
                    logger.debug("bjobs detected unfinished job %s", batchJobID)
                else:
                    status = self.boss.with_retries(
                        self._customGetJobExitCode, batchJobID, jobID
                    )
                    if status is not None:
                        activity = True
                        self.updatedJobsQueue.put(
                            UpdatedBatchJobInfo(
                                jobID=jobID,
                                exitStatus=status,
                                exitReason=None,
                                wallTime=None,
                            )
                        )
                        self.forgetJob(jobID)

            self._checkOnJobsCache = activity
            self._checkOnJobsTimestamp = datetime.now()
            return activity

        def _customGetJobExitCode(self, lsfID, jobID):
            """Get LSF exit code."""
            # the task is set as part of the job ID if using getBatchSystemID()
            if "." in lsfID:
                lsfID, _ = lsfID.split(".", 1)

            commands = [
                ["bjobs", "-l", str(lsfID)],
                ["bacct", "-l", str(lsfID)],
                ["bhist", "-l", "-n", "1", str(lsfID)],
                ["bhist", "-l", "-n", "2", str(lsfID)],
            ]

            for i in commands:
                logger.debug("Checking job via: %s", " ".join(i))
                status = self._processStatusCommandLSF(i, jobID)

                if status != self._CANT_DETERMINE_JOB_STATUS:
                    return status

            logger.debug("Can't determine status for job: %s", lsfID)
            return None

        def _processStatusCommandLSF(self, command, jobID):
            output = subprocess.check_output(command).decode("utf-8")
            cmdstr = " ".join(command)

            if "Done successfully" in output:
                logger.debug("Detected completed job: %s", cmdstr)
                status = 0

            elif "Completed <done>" in output:
                logger.debug("Detected completed job: %s", cmdstr)
                status = 0

            elif "TERM_MEMLIMIT" in output:
                status = self._customRetry(jobID, term_memlimit=True)

            elif "TERM_RUNLIMIT" in output:
                status = self._customRetry(jobID, term_runlimit=True)

            elif "New job is waiting for scheduling" in output:
                logger.debug("Detected job pending scheduling: %s", cmdstr)
                status = None

            elif "PENDING REASONS" in output:
                logger.debug("Detected pending job: %s", cmdstr)
                status = None

            elif "Started on " in output:
                logger.debug("Detected job started but not completed: %s", cmdstr)
                status = None

            elif "Completed <exit>" in output:
                logger.error("Detected failed job: %s", cmdstr)
                status = 1

            elif "Exited with exit code" in output:
                logger.error("Detected failed job: %s", cmdstr)
                status = 1

            else:
                status = self._CANT_DETERMINE_JOB_STATUS

            return status

        def _customRetry(self, jobID, term_memlimit=False, term_runlimit=False):
            """Retry job if killed by LSF due to runtime or memlimit problems."""
            try:
                jobNode = self.boss.Id2Node[jobID]
            except KeyError:
                logger.error("Can't resource retry %s, jobNode not found", jobID)
                return 1

            retry_type = "memlimit" if term_memlimit else "runlimit"
            jobNode.jobName = (jobNode.jobName or "") + " resource retry " + retry_type
            memory = MAX_MEMORY if term_memlimit else jobNode.memory
            runtime = MAX_RUNTIME if term_runlimit else None
            bsub_line = self.prepareBsub(jobNode.cores, memory, jobID, runtime)

            if retry_type not in self.boss.resourceRetryCount[jobID]:
                lsfID = self.submitJob((bsub_line + [jobNode.command], None))
                self.batchJobIDs[jobID] = (lsfID, None)
                self.boss.resourceRetryCount[jobID].add(retry_type)
                logger.info("Detected job killed by LSF, attempting retry: %s", lsfID)
            else:
                logger.error("Can't retry for %s twice: %s", retry_type, jobID)
                return 1

            return None

        @staticmethod
        def _getNotFinishedIDs():
            return {
                int(i)
                for i in subprocess.check_output(["bjobs", "-o", "id"])
                .decode("utf-8")
                .strip()
                .split("\n")[1:]
            }



"""toil_container jobs."""

import os
import logging
import subprocess

import coloredlogs
from slugify import slugify
from toil.job import Job
from toil.statsAndLogging import StatsAndLogging

from toil.batchSystems import registry
from toil_container import containers, exceptions
from toil_container.lsf import CustomLSFBatchSystem
from toil_container.lsf_helper import encode_dict


# add timestamps to logging and nice colors
logging_format = (  # pylint: disable=invalid-name
    "%(asctime)s %(name)s %(hostname)s [pid %(process)d]\n%(levelname)s:%(message)s\n"
)

try:
    coloredlogs.install(fmt=logging_format)
except:  # pylint: disable=bare-except
    pass


def logWithFormatting(  # pylint: disable=invalid-name
    jobStoreID, jobLogs, method=logging.getLogger(__name__).debug, message=None
):
    """Join job logs in a single log, it's much more readable."""
    if message is not None:
        method(message)
    if isinstance(jobStoreID, bytes):
        jobStoreID = jobStoreID.decode("utf-8")

    lines = ""

    for line in jobLogs:
        if isinstance(line, bytes):
            line = line.decode("utf-8")

        lines += "\t" + line

    method("Received logs from jobStoreID %s:\n\n%s", jobStoreID, lines.rstrip())


# overwrite
StatsAndLogging.logWithFormatting = staticmethod(logWithFormatting)

# register the custom LSF Batch System
registry.addBatchSystemFactory("custom_lsf", lambda: CustomLSFBatchSystem)


class ContainerJob(Job):

    """A job class with a `call` method for containerized system calls."""

    def __init__(self, options, runtime=None, *args, **kwargs):
        """
        Set toil's namespace `options` as an attribute.

        Note that `runtime (-W)` is custom LSF solutions that is ignored unless
        toil is run with `--batchSystem custom_lsf`. Please note that this hack
        encodes the requirements in the job's `unitName` resulting in longer
        log files names.

        Let us know if you need more custom parameters, e.g. `runtime_limit`,
        or if you know of a better solution (see: BD2KGenomics/toil#2065).

        Arguments:
            runtime (int): estimated run time for the job in minutes,
                ignored unless batchSystem is set to custom_lsf (-W).
            options (object): an `argparse.Namespace` object with toil options.
            args (list): positional arguments to be passed to `toil.job.Job`.
            kwargs (dict): key word arguments to be passed to `toil.job.Job`.
        """
        self.options = options

        if not kwargs.get("displayName"):
            kwargs["displayName"] = self.__class__.__name__

        if getattr(options, "batchSystem", None) == "custom_lsf":
            data = {"runtime": runtime or os.getenv("TOIL_CONTAINER_RUNTIME")}
            kwargs["unitName"] = str(kwargs.get("unitName", "") or "")
            kwargs["unitName"] += encode_dict(data)

        super().__init__(*args, **kwargs)

        # set jobName to displayName so that logs are named with displayName
        self.jobName = slugify(kwargs["displayName"], separator="_")

    def call(self, args, cwd=None, env=None, check_output=False):
        """
        Make a containerized call if images available, else use subprocess.

        Docker will be used if `self.options.docker` is set. Similarly,
        Singularity will be used if `self.options.singularity` is set.
        If neither case, `subprocess` will be used.

        If `self.options.workDir` is defined, this path will be used as the
        temporary directory within the containers.

        If `self.options.volumes` is available, these will be mounted
        inside the containers. `volumes` must be a list of tuples:

            [(<local_path>, <container_absolute_path>), ...]

        Arguments:
            args (list): list of command line arguments passed to the tool.
            cwd (str): current working directory.
            env (dict): environment variables to set inside container.
            check_output (bool): if true, returns stdout of system call.

        Returns:
            str: (check_output=True) stdout of the system call.
            int: (check_output=False) 0 if call succeed else raise error.

        Raises:
            toil_container.SystemCallError: if system call cannot be completed.
            toil_container.UsageError: if both `singularity` and `docker` are
                set in `self.options`. Or if invalid `volumes` are defined.
        """
        call_kwargs = dict(args=args, env=env, cwd=cwd)
        docker = getattr(self.options, "docker", None)
        singularity = getattr(self.options, "singularity", None)

        if singularity and docker:
            raise exceptions.UsageError("use docker or singularity, not both.")

        if singularity or docker:
            call_kwargs["check_output"] = check_output

            # used for testing only
            call_kwargs["remove_tmp_dir"] = getattr(self, "_rm_tmp_dir", True)

            if getattr(self.options, "workDir", None):
                call_kwargs["working_dir"] = self.options.workDir

            if getattr(self.options, "volumes", None):
                call_kwargs["volumes"] = self.options.volumes

            if singularity:
                call_kwargs["image"] = singularity
                call_function = containers.singularity_call
            else:
                call_kwargs["image"] = docker
                call_function = containers.docker_call

        elif check_output:
            call_function = subprocess.check_output

        else:
            call_function = subprocess.check_call

        errors = (exceptions.ContainerError, subprocess.CalledProcessError, OSError)

        try:
            output = call_function(**call_kwargs)
        except errors as error:  # pylint: disable=catching-non-exception
            raise exceptions.SystemCallError(error)

        try:
            return output.decode()
        except AttributeError:
            return output
